unit QzFtp.Defines;

// QzFtp.Defines.pas
// Ftp support utils and constants.

//Copyright (C) Joseph Leung 2005 (author2004 (.at.) quickzip (.dot.) org)
//
//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.
//
//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//Lesser General Public License for more details.
//
//You should have received a copy of the GNU Lesser General Public
//License along with this library; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

interface
uses
  System.Windows.Forms, System.ComponentModel, System.Collections, System.Net,
  System.Net.Sockets, System.Io, System.Text;
type
  aFtpMessageInfo = record
      Code   : integer;
      Desc   : string;
   end;
  const MaxBuiltinFtpMessages = 39;
  const BuiltinFtpMessages: array[1..MaxBuiltinFtpMessages] of aFtpMessageInfo = (
  {$REGION 'Built In FtpError List'}
  //500 Series: The command was not accepted and the requested action did not take place.
  (Code: 500; Desc: 'Syntax error, command unrecognized. This may include errors such as command line too long.'),
  (Code: 501; Desc: 'Syntax error in parameters or arguments.'),
  (Code: 502; Desc: 'Command not implemented.'),
  (Code: 503; Desc: 'Bad sequence of commands.'),
  (Code: 504; Desc: 'Command not implemented for that parameter.'),
  (Code: 530; Desc: 'Not logged in.'),
  (Code: 532; Desc: 'Need account for storing files.'),
  (Code: 550; Desc: 'Requested action not taken. File unavailable.'),
  (Code: 551; Desc: 'Requested action aborted. Page type unknown.'),
  (Code: 552; Desc: 'Requested file action aborted. Exceeded storage allocation.'),
  (Code: 553; Desc: 'Requested action not taken. File name not allowed.'),
  //400 Series: The command was not accepted and the requested action did not take place, but the error condition is temporary and the action may be requested again.
  (Code: 421; Desc: 'Service not available, closing control connection.This may be a reply to any command if the service knows it must shut down'),
  (Code: 425; Desc: 'Can''t open data connection.'),
  (Code: 426; Desc: 'Connection closed; transfer aborted.'),
  (Code: 450; Desc: 'Requested file action not taken.'),
  (Code: 451; Desc: 'Requested action aborted. Local error in processing.'),
  (Code: 452; Desc: 'Requested action not taken. Insufficient storage space in system.File unavailable.'),
  //300 Series: The command has been accepted, but the requested action is dormant, pending receipt of further information.
  (Code: 331; Desc: 'User name okay, need password.'),
  (Code: 332; Desc: 'Need account for login'),
  (Code: 350; Desc: 'Requested file action pending further information'),
  //200 Series: The requested action has been successfully completed.
  (Code: 200; Desc: 'Command okay.'),
  (Code: 202; Desc: 'Command not implemented, superfluous at this site.'),
  (Code: 211; Desc: 'System status, or system help reply.'),
  (Code: 212; Desc: 'Directory status'),
  (Code: 213; Desc: 'File status.'),
  (Code: 214; Desc: 'Help message.On how to use the server or the meaning of a particular non-standard command. This reply is useful only to the human user.'),
  (Code: 215; Desc: 'NAME system type. Where NAME is an official system name from the list in the Assigned Numbers docume'),
  (Code: 220; Desc: 'Service ready for new user.'),
  (Code: 221; Desc: 'Service closing control connection'),
  (Code: 225; Desc: 'Data connection open; no transfer in progress.'),
  (Code: 226; Desc: 'Closing data connection. Requested file action successful.'),
  (Code: 227; Desc: 'Entering Passive Mode (h1,h2,h3,h4,p1,p2).'),
  (Code: 230; Desc: 'User logged in, proceed. Logged out if appropriate.'),
  (Code: 250; Desc: 'Requested file action okay, completed.'),
  (Code: 257; Desc: '"PATHNAME" created.'),
  //100 Series: The requested action is being initiated, expect another reply before proceeding with a new command.
  (Code: 110; Desc: 'Restart marker reply. In this case, the text is exact and not left to the particular implementation.'),
  (Code: 120; Desc: 'Service ready in nnn minutes.'),
  (Code: 125; Desc: 'Data connection already open; transfer starting.'),
  (Code: 150; Desc: 'File status okay; about to open data connection.')
  {$ENDREGION}
  );
  type
  FtpMessageEventArgs = class(EventArgs)
    private
      pMessageCode : integer;
      pMessage : string;
      pCommand : string;
      pClient2Server : boolean;
      function getCmdCorrect : boolean;
      function getCmdAccepted : boolean;
      function getCmdTakePlace : boolean;
      function getMessageDesc : string;
      function getDefMessageDesc : string;
      function getServer2Client : boolean;
    public
      constructor Create(anErrorCode : integer; aMessage : string); overload;
      constructor Create(aCommand : string); overload;
      property MessageCode : integer read pMessageCode;
      property MessageDesc : string read getMessageDesc;
      property defMessafeDesc : string read getDefMessageDesc;
      property Accepted : boolean read getCmdAccepted;
      property TakingPlace : boolean read getCmdTakePlace;
      property CorrectSyntax : boolean read getCmdCorrect;
      property Client2Server : boolean read pClient2Server;
      property Server2Client : boolean read getServer2Client;
    end;

  FtpMessageHandler = procedure (Sender:System.Object; e : FtpMessageEventArgs) of object;
  FileContentType = class(System.Object)
    private
      pFilename   : string;
      pFilesize   : integer;
      pOwner      : string;
      pAttribute  : string;
      pFileDate   : Datetime;
      pCurrentDir : string;
    public
      constructor Create(input,currentDir : string); overload;
      function isFolder : boolean;
      function isFile : boolean;
      property Filename   : string read pFilename;
      property Filesize   : integer read pFilesize;
      property Owner      : string read pOwner;
      property Attribute  : string read pAttribute;
      property FileDate   : Datetime read pFileDate;
      property CurrentDir : string read pCurrentDir;
    end;

function AppendBackSlash(Input: string): string;
function AppendFrontBackSlash(Input: string): string;
function DecodeIPPortString(IPStr : string;var ip1,ip2,ip3,ip4,p1,p2 : integer) : boolean;

implementation


function AppendBackSlash(Input: string): string;
begin
  if Length(Input) > 0 then
    if Input[Length(Input)] = '/' then
      Result := Input
    else
      Result := Input + '/'
  else
    Result := Input;
end;

function AppendFrontBackSlash(Input: string): string;
begin
  if input = '' then result := '/' else
    if input[1] <> '/' then Result := '/' + input else
      Result := input;
end;

constructor FileContentType.Create(input,currentDir : string);
var count1,count2 : integer;
  function next4CharIsSpace(count : integer) : boolean;
  begin
    Result :=
      (input[count+0] = ' ') and
      (input[count+1] = ' ') and
      (input[count+2] = ' ') and
      (input[count+3] = ' ');
  end;
  function str2DateTime(str : string) : Datetime;
  // MMM_DD_hh:mm = 1
  // MMM_DD__YYYY = 2
  // MMM_DD_YYYY_ = 3
  var datetype : integer;
      month : string;
      day : string;
      year : string;
      hour : string;
      minute : string;
  begin
    if str.IndexOf(':') > 1 then
      dateType := 1 else
      if str[8] = ' ' then
        dateType := 3 else
        dateType := 2;

    year := Datetime.Now.Year.toString;
    Case dateType of
    2 : year := copy(str,9,4);
    3 : year := copy(str,8,4);
    end;
    month := Copy(str,1,3);
    day := Copy(str,5,2);

    if dateType = 1 then
      begin
        hour := Copy(str,8,2);
        minute := Copy(str,11,2);
      end;

    Case dateType of
    1 : Result := Convert.ToDateTime(Year+'/'+Month+'/'+Day+' '+Hour+':'+Minute+':00');
    2,3 : Result := Convert.ToDateTime(Year+'/'+Month+'/'+Day);
    end;
  end;
begin
  inherited Create;
  //Console.WriteLine(input);
  
  pCurrentDir := AppendBackSlash(CurrentDir);
  if length(input) > 10 then
  begin
    pAttribute := Copy(input,1,10);

    count1 := 2;

    while (input.Length - count1 > 4) and
      not next4CharIsSpace(count1) do
        Inc(count1);
    while (input.Length - count1 > 4) and
      not System.Char.isDigit(input[count1]) do
        Inc(count1);

    pFilesize := -1;

    if input.Length - count1 > 4 then
      begin
        count2 := input.IndexOf(' ',count1);
        try
          pFileSize := Convert.ToInt32(Copy(input,count1,count2-count1+1));
        except
          on ex : System.FormatException do pFileSize := -1;
        end;
      end;

    count1 := count2 + 1;
    count2 := count1 + 12;

    try
      pFileDate := str2DateTime(Copy(input,count1+1,count2-count1));
    except
      on ex : System.FormatException do pFileDate := Datetime.Now;
    end;

    count1 := count2 + 1;
    count2 := input.Length;
    pFilename := Copy(input,count1+1,count2-count1);
  end;
end;

function FileContentType.isFolder : boolean;
begin
  Result := false;
  if pAttribute <> '' then
    Result := (pAttribute[1] = 'd');
end;

function FileContentType.isFile : boolean;
begin
  Result := false;
  if pAttribute <> '' then
    Result := (pAttribute[1] = '-');
end;

function FtpMessageEventArgs.getCmdCorrect : boolean;
begin
  Result := (MessageCode < 500);
end;

function FtpMessageEventArgs.getCmdAccepted : boolean;
begin
  Result := (MessageCode < 400);
end;

function FtpMessageEventArgs.getCmdTakePlace : boolean;
begin
  Result := (MessageCode < 300);
end;

function FtpMessageEventArgs.getMessageDesc : string;
begin
  if Client2Server then Result := pCommand else result := pMessage;
end;

function FtpMessageEventArgs.getDefMessageDesc : string;
var i : integer;
begin
  Result := pMessage;
  if Client2Server then Result := pCommand else
  for i := 1 to MaxBuiltinFtpMessages do
    if MessageCode = BuiltInFtpMessages[i].Code then
      begin
        Result := BuiltInFtpMessages[i].Desc;
        exit;
      end;
end;

constructor FtpMessageEventArgs.Create(anErrorCode : integer; aMessage : string);
begin
  inherited Create;
  pMessage := aMessage;
  pCommand := '';
  pMessageCode := anErrorCode;
  pClient2Server := false;
end;

function FtpMessageEventArgs.getServer2Client : boolean;
begin
  result := not Client2Server;
end;

constructor FtpMessageEventArgs.Create(aCommand : string);
begin
  inherited Create;
  pCommand := aCommand;
  pClient2Server := true;
  pMessageCode := 0;
end;

function DecodeIPPortString(IPStr : string;var ip1,ip2,ip3,ip4,p1,p2 : integer) : boolean;
var startPos, endPos, seperatorPos : integer;
    IPString : string;
  function PopAnIP : string;
  begin
    seperatorPos := pos(',',IPString);
    if seperatorPos > 0 then
    begin
      result := Copy(IPString,1,seperatorPos-1);
      IPString := Copy(IPString,seperatorPos+1,length(IPString)-SeperatorPos);
    end else
    begin
      result := Copy(IPString,1,length(IPString)-1);
      IPString := '';
    end;
  end;
begin
  startPos := pos('(',IPStr);
  endPos   := pos(')',IPStr);
  result := false;

  if (startPos > 1) then
    if (endPos > 1) then
      begin
        IPString := Copy(IPStr,StartPos+1,endPos-StartPos);

        try
          ip1 := Convert.ToInt32(PopAnIP);
          ip2 := Convert.ToInt32(PopAnIP);
          ip3 := Convert.ToInt32(PopAnIP);
          ip4 := Convert.ToInt32(PopAnIP);
          p1  := Convert.ToInt32(PopAnIP);
          p2  := Convert.ToInt32(PopAnIP);
        except
          on ex : FormatException do Result := false;
        end;
        Result := true;
      end;
end;

end.
