{$Include defines.inc}
unit QzFtp.FtpClient;

// QzFtp.FtpClient.pas
// FtpClient for dotNet environment, support PPC.

//Copyright (C) Joseph Leung 2005 (author2004 (.at.) quickzip (.dot.) org)
//
//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.
//
//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//Lesser General Public License for more details.
//
//You should have received a copy of the GNU Lesser General Public
//License along with this library; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

interface
uses
  System.Windows.Forms, System.ComponentModel, System.Collections, System.Net,
  System.Net.Sockets, System.Io, System.Text, QzFtp.Defines;

type
  QzFtpClient = class(System.Object)
  private
    const LineBreak = #10#13;
    var
    pControlClient  : TcpClient;
    pControlSender  : StreamWriter;
    pControlReader  : StreamReader;
    pTransferMode   : Integer;

    pRemoteHost     : string;
    pRemotePort     : integer;
    pWorking        : boolean;
    pConnected      : boolean;
    pTimeout        : integer;
    pLastOutput     : string;
    pUser           : string;
    pPassword       : string;
    pPassive        : boolean; 
    pFtpHandler     : FtpMessageHandler;

    procedure outConsole(msg : string);

    function createDataSocket : Socket;
    function login(user, password : string) : boolean;

    procedure setTransferMode(value : integer);
    procedure setRemoteHost(value : string);
    procedure setRemotePort(value : integer);
    procedure setUser(value : string);
    procedure setPassword(value : string);
    procedure SetTimeout(value : integer);

    function HandleFtpMessage(ftpResponseCode : integer; ftpResponseMsg : string; expectedCode : array of integer) : boolean;
  public
    const ASCII   = 0;    // <-- TransferMode
          Binary  = 1;
    var
    lastResponseCode: integer;
    constructor Create; overload;
    function connect : boolean;
    procedure disconnect;
    function list(dir : string) : ArrayList;
    procedure get(filename : string;var OutputStream : Stream);
    procedure put(inputStream : Stream; remoteFilename : string; appendFile : boolean);

    function sendCommand(command : string; var Output : string; expectedCode : array of integer) : boolean; overload;
    function sendCommand(command : string; var Output : string) : integer; overload;
    function sendCommand(command : string) : integer; overload;
    function getResponse(var output : string) : integer; overload;
    function getResponse(var output : string; expectedCode : array of integer) : boolean; overload;

    property Connected : boolean read pConnected;
    property RemoteHost : string read pRemoteHost write setRemoteHost;
    property RemotePort : integer read pRemotePort write setRemotePort;
    property Timeout : integer read pTimeout write setTimeout;
    property User : string read pUser write setUser;
    property Password : string read pPassword write setPassword;
    property LastOutput : string read pLastOutput;
    property TransferMode : integer read pTransferMode write setTransfermode;
    property OnFtpMessage : FtpMessageHandler read pFtpHandler write pFtpHandler;
    property passiveMode : boolean read pPassive;
  end;

implementation



constructor QzFtpClient.Create;
begin
  inherited;
  pRemoteHost := 'localhost';
  pRemotePort := 21;
  pTimeout    := 10000;
  pUser       := 'anonymous';
  pPassword   := 'guest@anonymous.com';
  pWorking := false;
  pConnected := false;
end;

procedure QzFtpClient.outConsole(msg : string);
begin
  pLastOutput := msg;
  //Console.Writeline(msg);
end;

function QzFtpClient.sendCommand(command : string) : integer;
var output : string;
begin
  result := sendCommand(command,output);
end;

function QzFtpClient.sendCommand(command : string; var Output : string) : integer;
begin
  sendCommand(command,output,[]);
  Result := lastResponseCode;
end;

function QzFtpClient.sendCommand(command : string; var Output : string; expectedCode : array of integer) : boolean;
var ftpMsg : FtpMessageEventArgs;
    rCode : integer;
begin
  if pWorking then
    raise Exception.Create('Busy');
  if (pControlClient = nil) or not pConnected then
    raise Exception.Create('Not Connected');

  pWorking := true;
  Result := false;

  try
    if Assigned(pFtpHandler) then
    begin
    if copy(command,1,5) = 'PASS ' then
      ftpMsg := FtpMessageEventArgs.Create('PASS ****') else
      ftpMsg := FtpMessageEventArgs.Create(Command);
    pFtpHandler(Self,ftpMsg);
    end;

     pControlSender.Write(command+LineBreak);
     pControlSender.Flush;

     rCode := getResponse(Output);
     Result := HandleFtpMessage(rCode,output,expectedCode);

     lastResponseCode := rCode;
  finally
    pWorking := false;
  end;
end;

function QzFtpClient.getResponse(var output : string; expectedCode : array of integer) : boolean;
var rCode : integer;
begin
  rCode := getResponse(output);
  Result := HandleFtpMessage(rCode,output,expectedCode);
end;

function QzFtpClient.getResponse(var output : string) : integer;
var input : string;
    response : StringBuilder;
    ftpMsg : FtpMessageEventArgs;
  function getResponseCode : integer;
  begin
    try
      Result := Convert.ToInt32(Copy(input,1,3));
    except
      on ex : System.FormatException do Result := -1;
    end;
  end;
  function getResponseMsg : string;
  begin
    if length(input) > 5 then
        Result := (Copy(input,5,Length(input)-4)).Trim;
  end;
begin
  Result := -1;

  try
    input :=  pControlReader.ReadLine;
  except
    on ex : IOException do input := 'Read error (Connection lost?)';
  end;


  while (input <> '') and (input[4] = '-') do
      begin
        ftpMsg := FtpMessageEventArgs.Create(getResponseCode,getResponseMsg);
        if Assigned(OnFtpMessage) then
          OnFtpMessage(Self,ftpMsg);

        input :=  pControlReader.ReadLine;

        if input = '' then
          raise Exception.Create('Unexpected no data received');
      end;

  output := getResponseMsg;
  result := getResponseCode;
end;

function Device_Is_Cradled : boolean;
var localEndPoint : IPEndPoint;
begin
  localEndPoint := IPEndPoint.Create(Dns.Resolve(Dns.GetHostName()).AddressList[0], 0);
  Result := (localEndPoint.Address.ToString = '192.168.55.101');
end;

function QzFtpClient.createDataSocket : Socket;
    function ActiveSocket(var s : socket) : boolean;
    var localHostEntry : IPHostEntry;
      localEndPoint  : IPEndPoint;
      ipAddr          : IPAddress;
      hostBytes      : Array of Byte;
      portBytes      : Array of Byte;
      cmd            : string;
      pPort          : integer;
      pAddress       : IPAddress;
      output         : string;
      rCode          : integer;
        procedure calPortBytes(port : integer);
        begin
          SetLength(portBytes,2);
          portBytes[0] := (port SHR 8);    // bits 1- 8
			    portBytes[1] := (port AND $FF);  // bits 9-16
        end;
        procedure calHostBytes(ipaddress : string);
        var i : integer;
            addr : string;
          function popAnIP : integer;
          var p : integer;
          begin
            p := pos('.',addr);
            result := Convert.ToInt32(Copy(addr,1,p-1));
            addr := Copy(addr,p+1,length(addr)-p);
          end;
        begin
          addr := ipaddress;
          SetLength(hostBytes,4);
          hostBytes[0] := popAnIP;
          hostBytes[1] := popAnIP;
          hostBytes[2] := popAnIP;
          hostBytes[3] := Convert.ToInt32(addr);
        end;
    begin
        result := false;
        s := Socket.Create(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
        localHostEntry := Dns.Resolve(Dns.GetHostName);
        localEndPoint := IPEndPoint.Create(localHostEntry.AddressList[0] , 0);
        s.Bind(localEndPoint);

        s.Listen(5);
        with ((s.LocalEndPoint) as IPEndPoint) do
          begin
            pPort := ((s.LocalEndPoint) as IPEndPoint).Port;
            pAddress := ((s.LocalEndPoint) as IPEndPoint).Address;

            //hostBytes := Address.GetAddressBytes();
            calHostBytes(Address.ToString);
            calPortBytes(pPort);

            cmd := System.String.Format('PORT {0},{1},{2},{3},{4},{5}',
      [       Convert.ToInt32(hostBytes[0]),
              Convert.ToInt32(hostBytes[1]),
              Convert.ToInt32(hostBytes[2]),
              Convert.ToInt32(hostBytes[3]),
              Convert.ToInt32(portBytes[0]),
              Convert.ToInt32(portBytes[1])]);

          sendCommand(cmd,output,[200]);
          end;
    end;

    function PassiveSocket(var s : socket) : boolean;
    var ip1,ip2,ip3,ip4,p1,p2 : integer;
      remoteEndPoint  : IPEndPoint;
      ipAddr          : IPAddress;
      pPort          : integer;
      pAddress       : string;                       
      output : string;
    begin
      result := false;
      if SendCommand('PASV',output,[227]) then
        if DecodeIPPortString(output,ip1,ip2,ip3,ip4,p1,p2) then
          begin
            pAddress := System.String.Format
                ('{0}.{1}.{2}.{3}',[ip1,ip2,ip3,ip4]);
            pPort := (p1 * 256) + p2;

            ipAddr := IPAddress.Parse(pAddress);
            remoteEndPoint := IPEndPoint.Create(ipAddr,pPort);

            s := Socket.Create(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            s.Connect(remoteEndPoint);
            
            result := true;
          end;
    end;
begin
  pPassive := PassiveSocket(Result);
  if not pPassive then    //PASV not supported
      ActiveSocket(Result);
end;

procedure QzFtpClient.put(inputStream : Stream; remoteFilename : string; appendFile : boolean);
var cmd : string;
    output : string;
    line : string;
    pDataSocket  : socket;
    pDataReaderA : StreamReader;  pDataWriterA : StreamWriter;
    pDataReaderI : BinaryReader;  pDataWriterI : BinaryWriter;
  procedure setupStream;
  begin
    Case pTransfermode of
    ASCII  : begin
              pDataReaderA := StreamReader.Create(inputStream);
              if not passiveMode then
                pDataSocket := pDataSocket.Accept;
              pDataWriterA := StreamWriter.Create(NetworkStream.Create(pDataSocket));
             end;
    Binary : begin
              pDataReaderI := BinaryReader.Create(inputStream);
              if not passiveMode then
                pDataSocket := pDataSocket.Accept;
              pDataWriterI := BinaryWriter.Create(NetworkStream.Create(pDataSocket));
             end;
    else exception.Create('Unrecognized Transfer mode')
    end;
  end;
  procedure putFile;
  const size = 4096;
  var buffer : array of byte;
      ReadCnt : integer;
      totalReadCnt : integer;
  begin
    totalReadCnt := 0;
    Case pTransfermode of
    ASCII  : begin
              line := pDataReaderA.ReadLine;
              while (line <> nil) and (line <> '') do
              begin
                pDataWriterA.WriteLine(line);
                line := pDataReaderA.ReadLine;
              end;
             end;
    Binary : begin
              SetLength(buffer,size);
              ReadCnt := pDataReaderI.Read(buffer,0,size);
              while (ReadCnt > 0) do
              begin
                Inc(totalReadCnt,readCnt);
                pDataWriterI.Write(buffer,0,readCnt);
                readCnt := pDataReaderI.Read(buffer,0,size);
              end;
              //Console.WriteLine(totalReadCnt.toString + ' sent');
             end;
    else exception.Create('Unrecognized Transfer mode')
    end;
  end;
  procedure closeStream;
  begin
  Case pTransfermode of
    ASCII  : begin
              pDataWriterA.Flush;
              pDataWriterA.Close;
              pDataReaderA.Close;
             end;
    Binary : begin
              pDataWriterI.Flush;
              pDataWriterI.Close;
              pDataReaderI.Close;
             end;
    else exception.Create('Unrecognized Transfer mode')
    end;
  end;
begin
  pDataSocket := createDataSocket;
  setTransferMode(pTransfermode);

  if appendFile then cmd := 'APPE ' else cmd := 'STOR ';

  if sendCommand(cmd+ ' '+AppendFrontBackSlash(remoteFilename),Output,[125,150]) then
    begin
      setupStream;
      putFile;
      closeStream;
    end;
end;

procedure QzFtpClient.get(filename : string;var OutputStream : Stream);
var cmd : string;
    output : string;
    line : string;
    pDataSocket  : socket;
    pDataReaderA : StreamReader;  pDataWriterA : StreamWriter;
    pDataReaderI : BinaryReader;  pDataWriterI : BinaryWriter;
  procedure setupStream;
  begin
    Case pTransfermode of
    ASCII  : begin
              pDataWriterA := StreamWriter.Create(OutputStream);
              if not passiveMode then
                pDataSocket := pDataSocket.Accept;
              pDataReaderA := StreamReader.Create(NetworkStream.Create(pDataSocket));
             end;
    Binary : begin
              pDataWriterI := BinaryWriter.Create(OutputStream);
              if not passiveMode then
                pDataSocket := pDataSocket.Accept;
              pDataReaderI := BinaryReader.Create(NetworkStream.Create(pDatasocket));
             end;
    else exception.Create('Unrecognized Transfer mode')
    end;
  end;
  procedure getFile;
  const size = 4096;
  var buffer : array of byte;
      ReadCnt : integer;
  begin
    Case pTransfermode of
    ASCII  : begin
              line := pDataReaderA.ReadLine;
              while (line <> nil) and (line <> '') do
              begin
                pDataWriterA.WriteLine(line);
                line := pDataReaderA.ReadLine;
              end;
             end;
    Binary : begin
              SetLength(buffer,size);
              ReadCnt := pDataReaderI.Read(buffer,0,size);
              while (ReadCnt > 0) do
              begin
                pDataWriterI.Write(buffer,0,readCnt);
                readCnt := pDataReaderI.Read(buffer,0,size);
              end;
             end;
    else exception.Create('Unrecognized Transfer mode')
    end;
  end;
  procedure closeStream;
  begin
  Case pTransfermode of
    ASCII  : begin
              pDataWriterA.Flush;
              pDataWriterA.Close;
              pDataReaderA.Close;
             end;
    Binary : begin
              pDataWriterI.Flush;
              pDataWriterI.Close;
              pDataReaderI.Close;
             end;
    else exception.Create('Unrecognized Transfer mode')
    end;
  end;
begin
  pDataSocket := createDataSocket;
  setTransferMode(pTransfermode);

  if sendCommand('RETR '+AppendFrontBackSlash(filename),Output,[125,150]) then
    begin
      setupStream;
      getFile;
      closeStream;
      getResponse(output,[226,250]);
    end;
end;

function QzFtpClient.list(dir : string) : ArrayList;
var cmd : string;
    output : string;
    line : string;
    responseCode : integer;
    pDataSocket  : socket;
    pDataReader  : StreamReader;
    ftpMsg : FtpMessageEventArgs;
begin
  Result := ArrayList.Create;

  if dir = '' then
    cmd := 'LIST *' else
    cmd := 'LIST ' + AppendBackSlash(dir)+'*';

  pDataSocket := createDataSocket;

  if sendCommand('TYPE A',output,[200]) then
    if sendCommand(cmd,output,[125,150,550]) then
      begin
         if not passiveMode then
              pDataSocket := pDataSocket.Accept;
          pDataReader := StreamReader.Create(NetworkStream.Create(pDataSocket));
          
        line := pDataReader.ReadLine;
        while (line <> nil) and (line <> '') do
          begin
            Result.Add(FileContentType.Create(line,dir));

            ftpMsg := FtpMessageEventArgs.Create(000,line);
            if Assigned(OnFtpMessage) then
              OnFtpMessage(Self,ftpMsg);

            line := pDataReader.ReadLine;
          end;
        getResponse(output,[226,250]);
      end;
end;

procedure QzFtpClient.setTimeout(value : integer);
begin
  if Connected then
    raise Exception.Create('Disconnect first');
  pTimeout := value;
end;

function QzFtpClient.connect : boolean;
var remoteHostEntry : IPHostEntry;
    ipAddr : IPAddress;
    remoteEndPoint : IPEndPoint;
    stream : NetworkStream;
    output : string;
    ftpMsg : FtpMessageEventArgs;
begin
    result := false;

    pConnected := false;

    try
      ipAddr := IPAddress.Parse(pRemoteHost);
    except
      on e : Exception do
        ipAddr := (Dns.Resolve(pRemoteHost)).AddressList[0];
    end;
    pControlClient := TcpClient.Create;

    try
      pControlClient.Connect(ipAddr,pRemotePort);

      {$IFNDEF CE_FRAMEWORK}
      pControlClient.ReceiveTimeout := pTimeout;
      pControlClient.SendTimeout := pTimeout;
      {$ENDIF}

      stream := pControlClient.GetStream;
      pControlSender := StreamWriter.Create(stream);
      pControlReader := StreamReader.Create(stream);

      getResponse(output,[220]);
    except
      on e : SocketException do exit;
      on e : Exception do exit;
    end;

    pConnected := true;

    if not login(pUser,pPassword) then disconnect else
      Result := true;
end;

procedure QzFtpClient.disconnect;
var output : string;
begin
  if pConnected then
    begin
      sendCommand('QUIT',output);
      pControlSender.Close;
      pControlReader.Close;
      pControlClient.Close;
    end;

  pControlSender := nil;
  pControlReader := nil;
  pControlClient := nil;

  pConnected := false;
end;

function QzFtpClient.login(user, password : string) : boolean;
var Output : string;
begin
  Result := false;
  if sendCommand('USER '+user,output,[331]) then
      if sendCommand('PASS '+password,output,[230]) then
          Result := true;
end;


procedure QzFtpClient.setRemoteHost(value : string);
begin
  if Connected then
    raise Exception.Create('Disconnect first');
  pRemoteHost := value;
end;

procedure QzFtpClient.setRemotePort(value : integer);
begin
  if Connected then
    raise Exception.Create('Disconnect first');
  pRemotePort := value;
end;

procedure QzFtpClient.setUser(value : string);
begin
  if Connected then
    raise Exception.Create('Disconnect first');
  pUser := value;
  if pUser = '' then pUser := 'anonymous';
end;

procedure QzFtpClient.setPassword(value : string);
begin
  if Connected then
    raise Exception.Create('Disconnect first');
  pPassword := value;
end;

procedure QzFtpClient.setTransferMode(value : integer);
var cmd,output : string;
begin
  if (value < ASCII) or (value > Binary) then
    raise exception.Create('Unrecognized Transfer mode');
  pTransferMode := value;
  if Connected then
    begin
      if value = ASCII then cmd := 'TYPE A' else cmd := 'TYPE I';
      SendCommand(cmd,output,[200])
    end;
end;

function QzFtpClient.HandleFtpMessage(ftpResponseCode : integer; ftpResponseMsg : string; expectedCode : array of integer) : boolean;
var ftpMsg : FtpMessageEventArgs;
  function matchCode : boolean;
  var i : integer;
  begin
    Result := false;
    if length(expectedCode) = 0 then Result := true else
    for i := 0 to Length(expectedCode) -1 do
      if expectedCode[i] = ftpResponseCode then
        begin
          result := true;
          exit;
        end;
  end;
begin
  Result := false;
  ftpMsg := FtpMessageEventArgs.Create(ftpResponseCode,ftpResponseMsg);

  if Assigned(OnFtpMessage) then
    OnFtpMessage(Self,ftpMsg);

  if matchCode then
      Result := true else
      raise Exception.Create(ftpResponseCode.ToString + ' ' +
        ftpMsg.MessageDesc);
end;


end.

